/*! For license information please see task.27e92a39a0803fb0ad0d.js.LICENSE.txt */
!function(e,t){"object"==typeof exports&&"object"==typeof module?module.exports=t():"function"==typeof define&&define.amd?define([],t):"object"==typeof exports?exports.Airflow=t():(e.Airflow=e.Airflow||{},e.Airflow.task=t())}(self,(()=>(document.addEventListener("DOMContentLoaded",(()=>{document.querySelectorAll(".js-ti-attr").forEach((e=>{const t=e.innerHTML;32===t.length&&moment(t,"YYYY-MM-DD").isValid()?e.innerHTML=`<time datetime="${t}">${t}</time>`:t.includes("http")&&(e.innerHTML=`<a href=${t}>${t}</a>`)}))})),{})));